import re
import sys
import copy
import matplotlib.pyplot as plt
import numpy as np
import networkx as nx
import gurobipy as gp
from gurobipy import GRB

G = nx.DiGraph()


iterate = np.array([0, 10.6, 2, 9]) #distance between ports
iterate3 = np.append(iterate,iterate) # Adding the distance vector above to itself two times as the service does three cycles back to back
iterate4 = np.append(iterate3,iterate)
distac = np.add.accumulate(iterate4) # Accumulating the distances to find find distances to all stops from start node

num_cyc = int(len(iterate4)/len(iterate)) #Number of cycles on the route

distanceMatrSimple = np.zeros((len(distac),len(distac)))  #  creating distance matrix with distance from all ports to all ports
for i in range(len(distac)):
    for j in range(len(distac)):
        tall = np.abs(distac[i]-distac[j])
        distanceMatrSimple[i][j] = tall
    



N = np.arange(len(distanceMatrSimple)) # The set N with the nodes/stops

Start = N[0] #first node in set N
Slutt = N[-1] # Last node in set N


Bat_alt = [0,1,2,3] # Set with alternative for number of battery modules taken onboard in each stop. Expandedn to 4 for Baronessen and five for Baronen

#range_bat = [0,10,20,30] #Tidevind
#range_bat = [0, 11.8,23.6,35.4, 47.2] #Baronessen                  Range for vessels with 500kWh/module available, with 0-3,4 and 5 battery modules for Tidevind, Baronessen and Baronen, respectively.
#range_bat = [0,8.2,16.4,24.6,32.8,41]  #Baronen

range_bat = [0,8,16,24] #Tidevind80%DOD
#range_bat = [0, 9.42, 18.84, 28.26, 37.68] #Baronessen80%DOD
#range_bat = [0, 6.53, 13.06, 19.59, 26.12, 32.65] #Baronen80%DOD           Range for vessels with 400kWh/module available, with 0-3,4 and 5 battery modules for Tidevind, Baronessen and Baronen, respectively.

cost_bat = np.array([0, 0.25, 0.55, 0.85, 1.15, 1.45 ])    # Cost of takeing on different numbers of battery modules, from 0-5 modules


A = [(i,j) for i in N for j in N if i<j and j!=0] # All possible arcs between i and j, excluding when i == j
S = {(i,j): distanceMatrSimple[i,j] for i,j in A if distanceMatrSimple[i,j] <=max(range_bat) } #Set with all arcs and distances between arcs. As the route goes one way, duplicates are removed. Only generates arcs shortes than the range of the alternative taking on most battery modules.


#Cost of establishing charging in each port
cost = [0]*len(N)


cost[0] = 1.5/(num_cyc*2)
cost[3] = cost[0]                      #0,3,4,7,8,10 and 11 represetn the same port
cost[4] = cost[0]
cost[7] = cost[0]
cost[8] = cost[0]                   #The cost is split on the number of stops that represent the same port
cost[10] = cost[0]
cost[11] = cost[0]
cost[1] = 1.7/num_cyc
cost[5] = cost[1]
cost[9] = cost[1]                   # 1,5 and 9 represent the same port
cost[2] = 1.9/num_cyc
cost[6] = cost[2]                   #2,6 and 10 represent the same port
cost[10] = cost[2]





arcsData = []
for i,j in S:
    arcsData.append((i,j,S[i,j])) # Rewriting S to use it with gurobi



G.add_edges_from(S)

#GUROBI
#The optimization model 

m = gp.Model()
m.update()
x = m.addVars(S, vtype = GRB.BINARY, name = 'x') # Se thesis for description of variables
y = m.addVars(N,Bat_alt, vtype = GRB.BINARY, name = 'y')
u = m.addVars(N, vtype = GRB.BINARY, name = 'u')
w = m.addVars(S, vtype = GRB.BINARY, name = 'w')

#batCap = m.addVar(name = 'BatCap', vtype = GRB.CONTINUOUS)



m.modelSense = GRB.MINIMIZE #minimizing the Objective function
m.setObjective(gp.quicksum((cost[i])*u[i] for i in N ) + gp.quicksum(y[a,b]*cost_bat[b] for a in N for b in Bat_alt)  ) # The objective function


start_const = []
for a in arcsData:
    if a[0] == Start:                   # prep for constraint 6.6
        start_const.append(a)

end_const = []
for a in arcsData:
    if a[1] == Slutt:                   #prep for constraint 6.7
        end_const.append(a)

m.addConstrs(w[i,j]-x[i,j] == 0 for i,j in S)
m.addConstrs(w[i,j] >= x[i,j]+u[i]-1 for i,j in S)          #Linearization contraints. 6.2 -6.4
m.addConstrs(w[i,j]<= x[i,j] for i,j in S)
m.addConstrs(w[i,j]<= u[i] for i,j in S)


m.addLConstr(gp.quicksum(x[a[0],a[1]] for a in start_const) == 1) # Constraint 6.6: Only one of the arcs going from the start node can be included in the solution
m.addLConstr(gp.quicksum(x[a[0],a[1]] for a in end_const) == 1) # Constraint 6.7: Only one of the arcs going to the end node can be included in the solution
m.addConstrs((gp.quicksum(y[i,b]*range_bat[b] for b in Bat_alt) >= x[i,j]*S[i,j]  for i,j in S )) #constraint 6.9
m.addConstrs(gp.quicksum(y[i,b] for b in Bat_alt)<=1 for i in N) #constraint 6.10

m.addConstrs((u[i]-u[i+(int(len(N)/num_cyc))]  == 0 for i in N[0:(len(N)-len(iterate))]) ) #constraint for cyclic service. constraint 6.14
m.addConstr(u[Start] == u[Slutt]) # The first and last node represent the same port and thus needs to have the same value for u


#Constraint 6.8: The sum in and out of a node has to be equal for all nodes except the start and the end node
#  In combination with the above constraints this leads to a node having one arc in and one arc out in use(a part of the solution) or no arcs at all in use
for v in G.nodes:
    if v not in [Start,Slutt]:

        expr1 = gp.quicksum(x[i,v] for i in G.predecessors(v))

        expr2 = gp.quicksum(x[v,j] for j in G.successors(v))

        m.addLConstr(expr1 - expr2 == 0)

m.optimize() 

print(m.display())

m.write('test.mps')
if m.Status != GRB.OPTIMAL:
    print('Optimization ended with status %d' % m.Status)
    sys.exit(0)

for v in m.getVars():
    print('%s %g' % (v.VarName, v.X))


print('The Objective value: %g' % m.ObjVal)

#PRINTING SOLUTION
#Printing the nodes in the solutions in vector form
result_nodes = []
for v in m.getVars():

    if v.X >= 0.5 and 'x' in v.VarName:
        result_nodes.append(v.varName)


mellomhold = []
for a in result_nodes:
    b = re.findall(r'\d+',a)
    mellomhold.extend(b)

results_vec = []
for b in mellomhold:
    if b.isdigit():
        results_vec.append(int(b))

results_vec = list(sorted(set(results_vec)))
print(results_vec)


battery_data = []
for v in m.getVars():
    if v.X >= 0.5 and 'y' in v.VarName:
        battery_data.append(v.VarName)

mellomhold_bat = []
for a in battery_data:
    b = re.findall(r'\d+',a)
    mellomhold_bat.extend(b)

results_vec_y = []
for b in mellomhold_bat:
    if b.isdigit():
        results_vec_y.append(int(b))

battery_station = results_vec_y[0::2]
batter_choice = results_vec_y[1::2]

for i in range(len(battery_station)):
    print('Port %s will supply vessel with battery %s' %(battery_station[i],batter_choice[i]) )
        
print('The nodes in the solution:',results_vec)





pos = nx.kamada_kawai_layout(G) # Visulizing all possible feasible connections between the nodes



options = {'node_color': 'cyan','node_size': 200,'width': 0.5}

#Drawing network with all possible feasible connections between nodes
plt.figure()
nx.draw(G,pos, with_labels=True, **options)

#Drawing the network with nodes and arcs in the solution
plt.figure()
nx.draw(G,pos, with_labels=True, **options)
shortPath = [(results_vec[i],results_vec[i-1]) for i in range(len(results_vec)-1,0,-1)]
nx.draw_networkx_edges(G,pos,edgelist=shortPath,width=5,edge_color='red')


print(m.Runtime) # printing the runtime of the script


plt.show()

